//----------------------------------------------------------------------------
//
// File:        disassemble.cpp
// Date:        23-Feb-1998
// Programmer:  Marc Rousseau
//
// Description: Simple disassembler for the TMS9900 processor
//
// Copyright (c) 1994-2000 Marc Rousseau, All Rights Reserved.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
//
// Revision History:
//
//----------------------------------------------------------------------------

#include <stdio.h>
#include <string.h>
#include "common.hpp"
#include "tms9900.hpp"

static int     bUseR = 1;
static char   *pBuffer;
static USHORT  ByteTable [ 256 + 1 ];

extern "C" sLookUp LookUp [ 16 ];

void AddReg ( int reg )
{
    *pBuffer = 'R';
    pBuffer += bUseR;
    if ( reg >= 10 ) {
        reg -= 10;
        *pBuffer++ = '1';
    }
    *pBuffer++ = ( char ) ( '0' + reg );
}

void AddDigit ( int num )
{
    if ( num >= 10 ) {
        num -= 10;
        *pBuffer++ = '1';
    }
    *pBuffer++ = ( char ) ( '0' + num );
}

void AddByte ( UCHAR data )
{
  * ( USHORT * ) pBuffer = ByteTable [ data ], pBuffer += 2;
}

void AddWord ( USHORT data )
{
  * ( USHORT * ) pBuffer = ByteTable [ data >> 8 ], pBuffer += 2;
  * ( USHORT * ) pBuffer = ByteTable [ data & 0x00FF ], pBuffer += 2;
}

void GetRegs ( USHORT op, USHORT data )
{
    int reg = op & 0x0F;
    int mode = ( op >> 4 ) & 0x03;
    switch ( mode ) {
        case 0 : AddReg ( reg );
                 break;
        case 2 : *pBuffer++ = '@';
                 *pBuffer++ = '>';
                 AddWord ( data );
                 if ( reg != 0 ) {
                     *pBuffer++ = '(';
                     AddReg ( reg );
                     *pBuffer++ = ')';
                 }
                 break;
        case 1 :
        case 3 : *pBuffer++ = '*';
                 AddReg ( reg );
                 if ( mode == 3 ) *pBuffer++ = '+';
                 break;
    }
}

void format_I ( USHORT opcode, USHORT arg1, USHORT arg2 )
{
    GetRegs ( opcode, arg1 );
    *pBuffer++ = ',';
    GetRegs (( USHORT ) ( opcode >> 6 ), arg2 );
}

void format_II ( USHORT opcode, USHORT PC, USHORT )
{
    if ( opcode == 0x1000 ) {
        strcpy ( pBuffer - 5, "NOP" );
    } else {
        char disp = ( char ) opcode;
        *pBuffer++ = '>';
        if ( opcode >= 0x1D00 )
            AddByte ( disp );
        else
            AddWord (( USHORT ) ( PC + disp * 2 ));
    }
}

void format_VI ( USHORT opcode, USHORT arg1, USHORT )
{
    GetRegs ( opcode, arg1 );
}

void format_III ( USHORT opcode, USHORT arg1, USHORT arg2 )
{
    format_VI ( opcode, arg1, arg2 );
    *pBuffer++ = ',';
    AddReg (( opcode >> 6 ) & 0xF );
}

void format_IV ( USHORT opcode, USHORT arg1, USHORT arg2 )
{
    format_VI ( opcode, arg1, arg2 );
    UCHAR disp = ( UCHAR ) (( opcode >> 6 ) & 0xF );
    *pBuffer++ = ',';
    AddDigit ( disp ? disp : 16 );
}

void format_V ( USHORT opcode, USHORT, USHORT )
{
    AddReg ( opcode & 0xF );
    *pBuffer++ = ',';
    AddDigit (( opcode >> 4 ) & 0xF );
}

void format_VIII ( USHORT opcode, USHORT arg1, USHORT )
{
    if ( opcode < 0x02A0 ) {
        AddReg ( opcode & 0x000F );
        *pBuffer++ = ',';
        *pBuffer++ = '>';
        AddWord ( arg1 );
    } else if ( opcode >= 0x02E0 ) {
        *pBuffer++ = '>';
        AddWord ( arg1 );
    } else
        AddReg ( opcode & 0x000F );
}

USHORT GetArgs ( USHORT PC, USHORT *ptr, sOpCode *op, USHORT opcode )
{
    USHORT arg1 = 0, arg2 = 0;
    int index = 1;
    switch ( op->format ) {
        case 1 :				// Two General Addresses
                 if (( opcode & 0x0030 ) == 0x0020 ) { arg1 = ptr [index++]; PC += ( USHORT ) 2; }
                 if (( opcode & 0x0C00 ) == 0x0800 ) { arg2 = ptr [index]; PC += ( USHORT ) 2; }
                 break;
        case 3 :				// Logical
        case 4 :				// CRU Multi-Bit
        case 6 :				// Single Address
        case 9 :				// XOP, MULT, & DIV
                 if (( opcode & 0x0030 ) == 0x0020 ) { arg1 = ptr [index]; PC += ( USHORT ) 2; }
                 break;
        case 8 :				// Immediate
                 arg1 = ptr [index];
                 if (( opcode < 0x02A0 ) || ( opcode >= 0x02E0 )) PC += ( USHORT ) 2;
                 break;
    }

    switch ( op->format ) {
        case 1 : format_I ( opcode, arg1, arg2 );	break;
        case 2 : format_II ( opcode, PC, arg2 );	break;
        case 3 : format_III ( opcode, arg1, arg2 );	break;
        case 4 : format_IV ( opcode, arg1, arg2 );	break;
        case 5 : format_V ( opcode, arg1, arg2 );	break;
        case 6 : format_VI ( opcode, arg1, arg2 );	break;
//      case 7 : format_VII ( opcode, arg1, arg2 );	break;
        case 8 : format_VIII ( opcode, arg1, arg2 );	break;
        case 9 : format_III ( opcode, arg1, arg2 );	break;
    }

    return PC;
}

void InitByteTable ()
{
    for ( unsigned  i = 0; i < SIZE ( ByteTable ) - 1; i++ ) {
        sprintf (( char * ) &ByteTable [i], "%02X", i );
    }
}

USHORT DisassembleASM ( USHORT PC, UCHAR *ptr, char *buffer )
{
    static bool initialized = false;
    if ( initialized == false ) {
        InitByteTable ();
        initialized = true;
    }

    pBuffer = buffer;
    AddWord ( PC );
    *pBuffer++ = ' ';
    if ( PC & 1 ) {
        strcpy ( pBuffer, "<-- Illegal value in PC" );
        return ( USHORT ) ( PC + 1 );
    }
    USHORT curOpCode = (( USHORT * ) ptr ) [0];
    PC += ( USHORT ) 2;
    sLookUp *lookup = &LookUp [ curOpCode >> 12 ];
    sOpCode *op = lookup->opCode;
    int retries = lookup->size;
    if ( retries ) {
        while (( curOpCode & op->mask ) != op->opCode ) {
            op++;
            if ( retries-- == 0 ) {
                strcpy ( pBuffer, "Invalid Op-Code" );
                return PC;
            }
        }
    }
    * (( ULONG * ) pBuffer ) = * ( ULONG * ) op->mnemonic, pBuffer += 4;
    if ( op->format != 7 ) {
        *pBuffer++ = ' ';
        PC = GetArgs ( PC, ( USHORT * ) ptr, op, curOpCode );
    }
    *pBuffer = '\0';
    return PC;
}
